__all__ = (
    'ContentFilterLevel', 'GuildFeature', 'HubType', 'MFA', 'MessageNotificationLevel', 'NsfwLevel', 'VerificationLevel'
)
import warnings

from scarletio import class_property, export

from ...bases import Preinstance as P, PreinstancedBase


class VerificationLevel(PreinstancedBase):
    """
    Represents Discord's verification level.
    
    Attributes
    ----------
    name : `str`
        The default name of the verification level.
    value : `int`
        The discord side identifier value of the verification level.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``VerificationLevel``) items
        Stores the predefined ``VerificationLevel``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The verification levels' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the verification levels.
    
    Every predefined verification level can be accessed as class attribute as well:
    
    +-----------------------+-----------+-------+
    | Class attribute name  | name      | value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | low                   | low       | 1     |
    +-----------------------+-----------+-------+
    | medium                | medium    | 2     |
    +-----------------------+-----------+-------+
    | high                  | high      | 3     |
    +-----------------------+-----------+-------+
    | extreme               | extreme   | 4     |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    low = P(1, 'low')
    medium = P(2, 'medium')
    high = P(3, 'high')
    extreme = P(4, 'extreme')



class ContentFilterLevel(PreinstancedBase):
    """
    Represents Discord's content filter level.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the content filter level.
    name : `str`
        The default name of the content filter level.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ContentFilterLevel``) items
        Stores the predefined content filter levels. This container is accessed when translating a Discord side
        identifier of a content filter level. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `type` = `int`
        The verification filter levels' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the content filter levels.
    
    Every predefined content filter level is also stored as a class attribute:
    
    +-----------------------+-----------+-------+
    | Class attribute name  | name      | value |
    +=======================+===========+=======+
    | disabled              | disabled  | 0     |
    +-----------------------+-----------+-------+
    | no_role               | no_role   | 1     |
    +-----------------------+-----------+-------+
    | everyone              | everyone  | 2     |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    disabled = P(0, 'disabled')
    no_role = P(1, 'no_role')
    everyone = P(2, 'everyone')



@export
class GuildFeature(PreinstancedBase):
    """
    Represents a ``Guild``'s feature.
    
    Attributes
    ----------
    value : `str`
        The Discord side identifier value of the guild feature.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``GuildFeature``) items
        Stores the predefined ``GuildFeature``-s.
    VALUE_TYPE : `type` = `str`
        The guild features' values' type.
    DEFAULT_NAME : `str` = `''`
        The default name of the guild features. Guild features have the same value as name, so at their case it is not
        applicable.
    
    Every predefined guild feature can be accessed as class attribute as well:
    
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | Class attribute names                 | name                                  | value                                     |
    +=======================================+=======================================+===========================================+
    | application_command_permissions_v2    | application command permissions v2    | APPLICATION_COMMAND_PERMISSIONS_V2        |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | animated_banner                       | animated banner                       | ANIMATED_BANNER                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | animated_icon                         | animated icon                         | ANIMATED_ICON                             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | announcement_channels                 | announcement channels                 | NEWS                                      |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | auto_moderation_enabled               | auto moderation enabled               | AUTO_MODERATION                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | banner                                | banner                                | BANNER                                    |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | burst_reactions                       | burst reactions                       | BURST_REACTIONS                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | channel_banners                       | channel banners                       | CHANNEL_BANNER                            |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | channel_highlights                    | channel highlights                    | CHANNEL_HIGHLIGHTS                        |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | channel_highlights_disabled           | channel highlights disabled           | CHANNEL_HIGHLIGHTS_DISABLED               |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | commerce                              | commerce                              | COMMERCE                                  |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | creator_store_page                    | creator store page                    | CREATOR_STORE_PAGE                        |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | creator_monetizable                   | creator monetizable                   | CREATOR_MONETIZABLE                       |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | creator_monetizable_disabled          | creator monetizable disabled          | CREATOR_MONETIZABLE_DISABLED              |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | creator_monetizable_restricted        | creator monetizable restricted        | CREATOR_MONETIZABLE_RESTRICTED            |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | creator_monetizable_premium_service   | creator monetizable premium service   | CREATOR_MONETIZABLE_WHITEGLOVE            |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | creator_monetizable_temporarily       | creator monetizable temporarily       | CREATOR_MONETIZABLE_PROVISIONAL           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | community                             | community                             | COMMUNITY                                 |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | developer_support_guild               | developer support guild               | DEVELOPER_SUPPORT_SERVER                  |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | discoverable                          | discoverable                          | DISCOVERABLE                              |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | discoverable_disabled                 | discoverable disabled                 | DISCOVERABLE_DISABLED                     |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | discoverable_enabled_before           | discoverable enabled before           | ENABLED_DISCOVERABLE_BEFORE               |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | embedded_activities_experiment        | embedded activities experiment        | EXPOSED_TO_ACTIVITIES_WTP_EXPERIMENT      |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | featurable                            | featurable                            | FEATURABLE                                |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | has_directory_entry                   | has directory entry                   | HAS_DIRECTORY_ENTRY                       |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | home_override                         | home override                         | GUILD_HOME_OVERRIDE                       |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | home_test                             | home test                             | GUILD_HOME_TEST                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | hub                                   | hub                                   | HUB                                       |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | internal_employee_only                | internal employee only                | INTERNAL_EMPLOYEE_ONLY                    |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | invite_splash                         | invite splash                         | INVITE_SPLASH                             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | invites_disabled                      | invites disabled                      | INVITES_DISABLED                          |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | linked_to_hub                         | linked to hub                         | LINKED_TO_HUB                             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | marketplaces_connection_roles         | marketplaces connection roles         | MARKETPLACES_CONNECTION_ROLES             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | member_list_disabled                  | member list disabled                  | MEMBER_LIST_DISABLED                      |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | monetization_enabled                  | monetization enabled                  | MONETIZATION_ENABLED                      |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | more_emoji                            | more emoji                            | MORE_EMOJI                                |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | more_sticker                          | more sticker                          | MORE_STICKERS                             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | new_thread_permissions                | new thread permissions                | NEW_THREAD_PERMISSIONS                    |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | onboarding                            | onboarding                            | GUILD_ONBOARDING                          |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | onboarding_ever_enabled               | onboarding evert enabled              | GUILD_ONBOARDING_EVER_ENABLED             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | onboarding_has_prompts                | onboarding has prompts                | GUILD_ONBOARDING_HAS_PROMPTS              |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | partnered                             | partnered                             | PARTNERED                                 |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | preview_enabled                       | preview enabled                       | PREVIEW_ENABLED                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | private_threads                       | private threads                       | PRIVATE_THREADS                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | public                                | public                                | PUBLIC                                    |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | public_disabled                       | public disabled                       | PUBLIC_DISABLED                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | raid_alerts_enabled                   | raid alerts enabled                   | RAID_ALERTS_ENABLED                       |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | relay_enabled                         | relay enabled                         | RELAY_ENABLED                             |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | role_icons                            | role icons                            | ROLE_ICONS                                |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | role_subscriptions_enabled            | role subscriptions enabled            | ROLE_SUBSCRIPTIONS_ENABLED                |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | role_subscription_purchasable         | role subscription purchasable         | ROLE_SUBSCRIPTIONS_AVAILABLE_FOR_PURCHASE |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | text_in_stage_enabled                 | text in stage enabled                 | TEXT_IN_STAGE_ENABLED                     |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | text_in_voice_enabled                 | text in voice enabled                 | TEXT_IN_VOICE_ENABLED                     |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | thread_archive_seven_day              | thread archive seven day              | SEVEN_DAY_THREAD_ARCHIVE                  |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | thread_archive_three_day              | thread archive three day              | THREE_DAY_THREAD_ARCHIVE                  |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | threads_enabled                       | threads enabled                       | THREADS_ENABLED                           |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | threads_enabled_testing               | threads enabled testing               | THREADS_ENABLED_TESTING                   |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | thread_limit_increased                | thread limit increased                | INCREASED_THREAD_LIMIT                    |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | ticket_events_enabled                 | ticket events enabled                 | TICKETED_EVENTS_ENABLED                   |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | vanity_invite                         | vanity invite                         | VANITY_URL                                |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | verified                              | verified                              | VERIFIED                                  |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | vip_voice_regions                     | vip voice regions                     | VIP_REGIONS                               |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | welcome_screen_enabled                | welcome screen enabled                | WELCOME_SCREEN_ENABLED                    |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    | verification_screen_enabled           | verification screen enabled           | MEMBER_VERIFICATION_GATE_ENABLED          |
    +---------------------------------------+---------------------------------------+-------------------------------------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    
    __slots__ = ()
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new guild feature with the given value.
        
        Parameters
        ----------
        value : `str`
            The guild feature's identifier value.
        
        Returns
        -------
        self : ``GuildFeature``
            The created guild feature.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value.lower().replace('_', ' ')
        self.INSTANCES[value] = self
        return self
    
    
    # predefined
    application_command_permissions_v2 = P('APPLICATION_COMMAND_PERMISSIONS_V2', 'application command permissions v2')
    animated_banner = P('ANIMATED_BANNER', 'animated banner')
    animated_icon = P('ANIMATED_ICON', 'animated icon')
    announcement_channels = P('NEWS', 'announcement channels')
    auto_moderation_enabled = P('AUTO_MODERATION', 'auto moderation enabled')
    banner = P('BANNER', 'banner')
    burst_reactions = P('BURST_REACTIONS', 'burst reactions')
    channel_banners = P('CHANNEL_BANNER', 'channel banners')
    channel_highlights = P('CHANNEL_HIGHLIGHTS', 'channel highlights')
    channel_highlights_disabled = P('CHANNEL_HIGHLIGHTS_DISABLED', 'channel highlights disabled')
    commerce = P('COMMERCE', 'commerce')
    creator_store_page = P('CREATOR_STORE_PAGE', 'creator store page')
    creator_monetizable = P('CREATOR_MONETIZABLE', 'creator monetizable')
    creator_monetizable_disabled = P('CREATOR_MONETIZABLE_DISABLED', 'creator monetizable disabled')
    creator_monetizable_restricted = P('CREATOR_MONETIZABLE_RESTRICTED', 'creator monetizable restricted')
    creator_monetizable_premium_service = P('CREATOR_MONETIZABLE_WHITEGLOVE', 'creator monetizable premium service')
    creator_monetizable_temporarily = P('CREATOR_MONETIZABLE_PROVISIONAL', 'creator monetizable temporarily')
    community = P('COMMUNITY', 'community')
    developer_support_guild = P('DEVELOPER_SUPPORT_SERVER', 'developer support guild')
    discoverable = P('DISCOVERABLE', 'discoverable')
    discoverable_disabled = P('DISCOVERABLE_DISABLED', 'discoverable disabled')
    embedded_activities_experiment = P('EXPOSED_TO_ACTIVITIES_WTP_EXPERIMENT', 'embedded activities experiment')
    discoverable_enabled_before = P('ENABLED_DISCOVERABLE_BEFORE', 'discoverable enabled before')
    featurable = P('FEATURABLE', 'featurable')
    has_directory_entry = P('HAS_DIRECTORY_ENTRY', 'has directory_entry')
    home_override = P('GUILD_HOME_OVERRIDE', 'home override')
    home_test = P('GUILD_HOME_TEST', 'home test')
    hub = P('HUB', 'hub')
    internal_employee_only = P('INTERNAL_EMPLOYEE_ONLY', 'internal employee only')
    invite_splash = P('INVITE_SPLASH', 'invite splash')
    invites_disabled = P('INVITES_DISABLED', 'invites disabled')
    linked_to_hub = P('LINKED_TO_HUB', 'linked to hub')
    marketplaces_connection_roles = P('MARKETPLACES_CONNECTION_ROLES', 'marketplaces connection roles')
    member_list_disabled = P('MEMBER_LIST_DISABLED', 'member list disabled')
    monetization_enabled = P('MONETIZATION_ENABLED', 'monetization enabled')
    more_emoji = P('MORE_EMOJI', 'more emoji')
    more_sticker = P('MORE_STICKERS', 'more sticker')
    new_thread_permissions = P('NEW_THREAD_PERMISSIONS', 'new thread permissions')
    onboarding = P('GUILD_ONBOARDING', 'onboarding')
    onboarding_ever_enabled  = P('GUILD_ONBOARDING_EVER_ENABLED', 'onboarding evert enabled')
    onboarding_has_prompts  = P('GUILD_ONBOARDING_HAS_PROMPTS', 'onboarding has prompts')
    partnered = P('PARTNERED', 'partnered')
    preview_enabled = P('PREVIEW_ENABLED', 'preview enabled')
    private_threads = P('PRIVATE_THREADS', 'private threads')
    public = P('PUBLIC', 'public')
    public_disabled = P('PUBLIC_DISABLED', 'public disabled')
    raid_alerts_enabled = P('RAID_ALERTS_ENABLED', 'raid alerts enabled')
    relay_enabled = P('RELAY_ENABLED', 'relay enabled')
    role_icons = P('ROLE_ICONS', 'role icons')
    role_subscriptions_enabled = P('ROLE_SUBSCRIPTIONS_ENABLED', 'role subscriptions enabled')
    role_subscription_purchasable = P('ROLE_SUBSCRIPTIONS_AVAILABLE_FOR_PURCHASE', 'role subscription purchasable')
    text_in_stage_enabled = P('TEXT_IN_STAGE_ENABLED', 'text in stage enabled')
    text_in_voice_enabled = P('TEXT_IN_VOICE_ENABLED', 'text in voice enabled')
    thread_archive_seven_day = P('SEVEN_DAY_THREAD_ARCHIVE', 'thread archive seven day')
    thread_archive_three_day = P('THREE_DAY_THREAD_ARCHIVE', 'thread archive three day')
    threads_enabled = P('THREADS_ENABLED', 'threads enabled')
    threads_enabled_testing = P('THREADS_ENABLED_TESTING', 'threads enabled testing')
    thread_limit_increased = P('INCREASED_THREAD_LIMIT', 'thread limit increased')
    ticket_events_enabled = P('TICKETED_EVENTS_ENABLED', 'ticket events enabled')
    vanity_invite = P('VANITY_URL', 'vanity invite')
    verification_screen_enabled = P('MEMBER_VERIFICATION_GATE_ENABLED', 'verification screen enabled')
    verified = P('VERIFIED', 'verified')
    vip_voice_regions = P('VIP_REGIONS', 'vip_voice_regions')
    welcome_screen_enabled = P('WELCOME_SCREEN_ENABLED', 'welcome screen enabled')


class NsfwLevel(PreinstancedBase):
    """
    Represents a guild's nsfw level.
    
    Attributes
    ----------
    name : `str`
        The name of the nsfw filter level.
    value : `int`
        The identifier value the nsfw filter level
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``NsfwLevel``) items
        Stores the predefined ``NsfwLevel``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The nsfw level' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the nsfw levels.
    
    Every predefined nsfw level can be accessed as class attribute as well:
    
    +-----------------------+-------------------+-------+-------+
    | Class attribute name  | Name              | Value | nsfw  |
    +=======================+===================+=======+=======+
    | none                  | none              | 0     | False |
    +-----------------------+-------------------+-------+-------+
    | explicit              | explicit          | 1     | True  |
    +-----------------------+-------------------+-------+-------+
    | safe                  | safe              | 2     | False |
    +-----------------------+-------------------+-------+-------+
    | age_restricted        | age_restricted    | 3     | True  |
    +-----------------------+-------------------+-------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ('nsfw', )
    
    def __init__(self, value, name, nsfw):
        """
        Creates a new nsfw level instance.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value of the nsfw level.
        name : `str`
            The nsfw level name.
        nsfw : `bool`
            Whether the nsfw level refers to being actually nsfw.
        """
        self.value = value
        self.name = name
        self.nsfw = nsfw
        
        self.INSTANCES[value] = self
    
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new nsfw level from the given value
        
        Parameters
        ----------
        value : `int`
            The nsfw level's identifier value.
        
        Returns
        -------
        self : ``NsfwLevel``
            The created nsfw level.
        """
        self = object.__new__(cls)
        
        self.value = value
        self.name = ''
        self.nsfw = True
        self.INSTANCES[value] = self
        
        return self
    
    
    def __repr__(self):
        """Returns the nsfw level's representation."""
        return f'{self.__class__.__name__}(value = {self.value!r}, name = {self.name!r}, nsfw = {self.nsfw!r})'
    
    
    none = P(0, 'none', False)
    explicit = P(1, 'explicit', True)
    safe = P(2, 'safe', False)
    age_restricted = P(3, 'age_restricted', True)


class MessageNotificationLevel(PreinstancedBase):
    """
    Represents the default message notification level of a ``Guild``.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the message notification level.
    name : `str`
        The default name of the message notification level.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``MessageNotificationLevel``) items
        Stores the predefined message notification levels. This container is accessed when translating message
        notification level's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The notification levels' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the notification levels.
    
    Each predefined message notification level can also be accessed as a class attribute:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | all_messages          | all_messages  | 0     |
    +-----------------------+---------------+-------+
    | only_mentions         | only_mentions | 1     |
    +-----------------------+---------------+-------+
    | no_message            | no_messages   | 2     |
    +-----------------------+---------------+-------+
    | none                  | none          | 3     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    all_messages = P(0, 'all_messages')
    only_mentions = P(1, 'only_mentions')
    no_messages = P(2, 'no_messages')
    none = P(3, 'none')
    
    
    @class_property
    def null(cls):
        """
        `.null` is deprecated and will be removed in 2023 January. Please use `.none` instead.
        """
        warnings.warn(
            (
                f'`{cls.__name__}.null` is deprecated and will be removed in 2023 December. '
                f'Please use `.none` instead.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        
        return cls.none


class MFA(PreinstancedBase):
    """
    Represents Discord's Multi-Factor Authentication's levels.
    
    Attributes
    ----------
    name : `str`
        The default name of the MFA level.
    value : `int`
        The Discord side identifier value of the MFA level.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``MFA``) items
        Stores the predefined MFA level. This container is accessed when converting an MFA level's value to
        it's wrapper side representation.
    VALUE_TYPE : `type` = `int`
        The mfa levels' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the mfa levels.
    
    Each predefined MFA can also be accessed as class attribute:
    
    +-----------------------+-----------+-------+
    | Class attribute name  | name      | value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | elevated              | elevated  | 1     |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # Predefined
    none = P(0, 'none')
    elevated = P(1, 'elevated')


class VerificationFieldPlatform(PreinstancedBase):
    """
    Represents the verification field platform of a verification screen.
    
    Attributes
    ----------
    name : `str`
        The name of the verification field platform.
    value : `str`
        The Discord side identifier value of the verification field platform.
    

    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``VerificationFieldPlatform``) items
        Stores the predefined ``VerificationFieldPlatform``-s.
    VALUE_TYPE : `type` = `str`
        The verification field platforms' values' type.
    DEFAULT_NAME : `str` = `''`
        The default name of the verification field platforms. Verification field platforms have their name generated from
        their value, so at their case it is not applicable.
    
    Every predefined verification field platform can be accessed as class attribute as well:
    
    +-----------------------+-----------+-----------+
    | Class attribute names | Name      | Value     |
    +=======================+===========+===========+
    | email                 | email     | email     |
    +-----------------------+-----------+-----------+
    | phone                 | phone     | phone     |
    +-----------------------+-----------+-----------+
    """
    
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    
    __slots__ = ()
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new verification field platform from the given value.
        
        Parameters
        ----------
        value : `str`
            The verification field platform's identifier value.
        
        Returns
        -------
        self : ``VerificationFieldPlatform``
            The verification field platform.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value
        self.INSTANCES[value] = self
        return self
    
    def __repr__(self):
        """Returns the representation of the verification field platform."""
        return f'{self.__class__.__name__}(value = {self.value!r})'
    
    email = P('email', 'email')
    phone = P('phone', 'phone')


class HubType(PreinstancedBase):
    """
    Represents Discord's guild's hub type.
    
    Attributes
    ----------
    name : `str`
        The default name of the hub type
    value : `int`
        The Discord side identifier value of the hub type
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``HubType``) items
        Stores the predefined Hub types. This container is accessed when converting an Hub types' value to
        it's wrapper side representation.
    VALUE_TYPE : `type` = `int`
        The hub types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the hub types
    
    Each predefined hub type can also be accessed as class attribute:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | high_school           | high school   | 1     |
    +-----------------------+---------------+-------+
    | college               | college       | 2     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # Predefined
    none = P(0, 'none')
    high_school = P(1, 'high school')
    college = P(2, 'college')



__all__ = ()

from .base import AutoModerationRuleTriggerMetadataBase
from .keyword import AutoModerationRuleTriggerMetadataKeyword
from .keyword_preset import AutoModerationRuleTriggerMetadataKeywordPreset
from .mention_spam import AutoModerationRuleTriggerMetadataMentionSpam


def try_get_auto_moderation_trigger_metadata_type_from_data(data):
    """
    Tries to detect what type of auto moderation trigger metadata the given data is.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Auto moderation trigger metadata data.
    
    Returns
    -------
    metadata_type : `type<AutoModerationRuleTriggerMetadataBase>`
    """
    if ('keyword_filter' in data) or ('regex_patterns' in data):
        metadata_type = AutoModerationRuleTriggerMetadataKeyword
    elif ('presets' in data):
        metadata_type = AutoModerationRuleTriggerMetadataKeywordPreset
    elif ('mention_limit' in data):
        metadata_type = AutoModerationRuleTriggerMetadataMentionSpam
    else:
        metadata_type = AutoModerationRuleTriggerMetadataBase
    
    return metadata_type




__all__ = ('MessageType', )

import warnings

from scarletio import class_property, include

from ...activity import ACTIVITY_TYPES
from ...bases import Preinstance as P, PreinstancedBase
from ...embed import EmbedType
from ...utils import elapsed_time, sanitize_mentions


Client = include('Client')


def MESSAGE_DEFAULT_CONVERTER(self):
    """
    Default message content converter.
    
    Parameters
    ----------
    self : ``Message``
        The message, what's content will be converted.
    
    Returns
    -------
    content : `None`, `str`
        The converted content if applicable. Might be empty string.
    """
    content = self.content
    if (content is not None):
        content = sanitize_mentions(content, self.guild)
    
    return content


def convert_user_add(self):
    return f'{self.author.name} added {self.mentioned_users[0].name} to the group.'

def convert_user_remove(self):
    return f'{self.author.name} removed {self.mentioned_users[0].name} from the group.'


def convert_call(self):
    call = self.call
    if call is None or not any(isinstance(user, Client) for user in call.iter_users()):
        return f'{self.author.name} started a call.'
    
    ended_at = call.ended_at
    if ended_at is None:
        # We are using a unicode as dash and not the minus sign
        return f'{self.author.name} started a call — Join the call.'
    
    # This formatting could be more accurate.
    if elapsed_time is None:
        return f'{self.author.name} started a call.'
    
    return f'{self.author.name} started a call that lasted {elapsed_time(self.created_at, ended_at)}.'


def convert_channel_name_change(self):
    return f'{self.author.name} changed the channel name: {self.content!s}'

def convert_channel_icon_change(self):
    return f'{self.author.name} changed the channel icon.'

def convert_new_pin(self):
    return f'{self.author.name_at(self.guild_id)} pinned a message to this channel. See all pinned messages.'


JOIN_MESSAGE_FORMATTERS = (
    lambda name : f'{name} just joined the server - glhf!',
    lambda name : f'{name} just joined. Everyone, look busy!',
    lambda name : f'{name} just joined. Can I get a heal?',
    lambda name : f'{name} joined your party.',
    lambda name : f'{name} joined. You must construct additional pylons.',
    lambda name : f'Ermagherd. {name} is here.',
    lambda name : f'Welcome, {name}. Stay awhile and listen.',
    lambda name : f'Welcome, {name}. We were expecting you ( ͡° ͜ʖ ͡°)',
    lambda name : f'Welcome, {name}. We hope you brought pizza.',
    lambda name : f'Welcome {name}. Leave your weapons by the door.',
    lambda name : f'A wild {name} appeared.',
    lambda name : f'Swoooosh. {name} just landed.',
    lambda name : f'Brace yourselves. {name} just joined the server.',
    lambda name : f'{name} just joined... or did they?',
    lambda name : f'{name} just arrived. Seems OP - please nerf.',
    lambda name : f'{name} just slid into the server.',
    lambda name : f'A {name} has spawned in the server.',
    lambda name : f'Big {name} showed up!',
    lambda name : f'Where’s {name}? In the server!',
    lambda name : f'{name} hopped into the server. Kangaroo!!',
    lambda name : f'{name} just showed up. Hold my beer.',
    lambda name : f'Challenger approaching - {name} has appeared!',
    lambda name : f'It\'s a bird! It\'s a plane! Nevermind, it\'s just {name}.',
    lambda name : f'It\'s {name}! Praise the sun! [T]/',
    lambda name : f'Never gonna give {name} up. Never gonna let {name} down.',
    lambda name : f'{name} has joined the battle bus.',
    lambda name : f'Cheers, love! {name}\'s here!',
    lambda name : f'Hey! Listen! {name} has joined!',
    lambda name : f'We\'ve been expecting you {name}',
    lambda name : f'It\'s dangerous to go alone, take {name}!',
    lambda name : f'{name} has joined the server! It\'s super effective!',
    lambda name : f'Cheers, love! {name} is here!',
    lambda name : f'{name} is here, as the prophecy foretold.',
    lambda name : f'{name} has arrived. Party\'s over.',
    lambda name : f'Ready player {name}',
    lambda name : f'{name} is here to kick butt and chew bubblegum. And {name} is all out of gum.',
    lambda name : f'Hello. Is it {name} you\'re looking for?',
    lambda name : f'{name} has joined. Stay a while and listen!',
    lambda name : f'Roses are red, violets are blue, {name} joined this server with you',
)

# TODO: this system changed, just pulled out the new texts from the js client source, but the calculation is bad
def convert_welcome(self):
    formatter = JOIN_MESSAGE_FORMATTERS[int(self.created_at.timestamp()) % len(JOIN_MESSAGE_FORMATTERS)]
    user_name = self.author.name_at(self.guild_id)
    return formatter(user_name)


def convert_guild_boost(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
    return f'{self.author.name} boosted {guild_name} with Nitro!'


def convert_guild_boost_tier_1(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
        
    return f'{self.author.name} boosted {guild_name} with Nitro! {guild_name} has achieved level 1!'


def convert_guild_boost_tier_2(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
    
    return f'{self.author.name} boosted {guild_name} with Nitro! {guild_name} has achieved level 2!'


def convert_guild_boost_tier_3(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
        
    return f'{self.author.name} boosted {guild_name} with Nitro! {guild_name} has achieved level 3!'


def convert_new_follower_channel(self):
    channel = self.channel
    guild = channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
    
    user_name = self.author.name_at(guild)
    
    return (
        f'{user_name} has added {guild_name} #{channel.name} to this channel. '
        f'Its most important updates will show up here.'
    )


def convert_stream(self):
    user = self.author
    for activity in user.activities:
        if activity.type == ACTIVITY_TYPES.stream:
            activity_name = activity.name
            break
    else:
        activity_name = 'Unknown'
    
    user_name = user.name_at(self.guild_id)
    
    return f'{user_name} is live! Now streaming {activity_name}'


def convert_discovery_disqualified(self):
    return (
        'This server has been removed from Server Discovery because it no longer passes all the requirements. '
        'Check `Server Settings` for more details.'
    )


def convert_discovery_requalified(self):
    return 'This server is eligible for Server Discovery again and has been automatically relisted!'


def convert_discovery_grace_period_initial_warning(self):
    return (
        'This server has failed Discovery activity requirements for 1 week. '
        'If this server fails for 4 weeks in a row, it will be automatically removed from Discovery.'
    )

def convert_discovery_grace_period_final_warning(self):
    return (
        'This server has failed Discovery activity requirements for 3 weeks in a row. '
        'If this server fails for 1 more week, it will be removed from Discovery.'
    )


def convert_thread_created(self):
    return f'{self.author.name_at(self.guild_id)} started a thread: {self.content!r}. See all threads.'


def convert_auto_moderation_action(self):
    embed = self.embed
    if embed is None or embed.type is not EmbedType.auto_moderation_message:
        return None
    
    content_parts = []
    should_add_next = ''
    
    description = embed.description
    if description is not None:
        content_parts.append(description)
        should_add_next = '\n'
    
    
    for field in embed.iter_fields():
        if field.name == 'keyword':
            content_parts.append(should_add_next)
            content_parts.append('Keyword: ')
            content_parts.append(field.value)
            should_add_next = ' • '
            break
    
    for field in embed.iter_fields():
        if field.name == 'rule_name':
            content_parts.append(should_add_next)
            content_parts.append('Rule: ')
            content_parts.append(field.value)
            should_add_next = ' • '
            break
    
    for field in embed.iter_fields():
        if field.name == 'timeout_duration':
            content_parts.append(should_add_next)
            content_parts.append('Time-out: ')
            content_parts.append(field.value)
            content_parts.append(' secs')
            break
    
    return ''.join(content_parts)


def convert_invite_reminder(self):
    return (
        'Wondering who to invite?\n'
        'Start by inviting anyone who can help you build the server!'
    )


def convert_stage_start(self):
    return f'{self.author.name_at(self.guild_id)} started {self.content!s}.'


def convert_stage_end(self):
    return f'{self.author.name_at(self.guild_id)} ended {self.content!s}.'


def convert_stage_topic_change(self):
    return f'{self.author.name_at(self.guild_id)} changed the Stage topic: {self.content!r}'


def convert_stage_speaker(self):
    return f'{self.author.name_at(self.guild_id)} is now a speaker'


class MessageType(PreinstancedBase):
    """
    Represents a ``Message``'s type.
    
    Attributes
    ----------
    name : `str`
        The default name of the message type.
    value : `int`
        The Discord side identifier value of the message type.
    converter : `FunctionType`
        The converter function of the message type, what tries to convert the message's content to it's Discord side
        representation.
    deletable : `bool`
        Whether the message with this type can be deleted.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``MessageType``) items
        Stores the predefined ``MessageType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The message types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the message types.
    
    Every predefined message type can be accessed as class attribute as well:
    
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | Class attribute name                      | Name                                      | Value | Converter                                         | Deletable |
    +===========================================+===========================================+=======+===================================================+===========+
    | default                                   | default                                   | 0     | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | user_add                                  | user add                                  | 1     | convert_user_add                                  | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | user_remove                               | user remove                               | 2     | convert_user_remove                               | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | call                                      | call                                      | 3     | convert_call                                      | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | channel_name_change                       | channel name change                       | 4     | convert_channel_name_change                       | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | channel_icon_change                       | channel icon change                       | 5     | convert_channel_icon_change                       | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | new_pin                                   | new pin                                   | 6     | convert_new_pin                                   | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | welcome                                   | welcome                                   | 7     | convert_welcome                                   | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost                               | guild boost                               | 8     | convert_guild_boost                               | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost_tier_1                        | guild boost tier 1                        | 9     | convert_guild_boost_tier_1                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost_tier_2                        | guild boost tier 2                        | 10    | convert_guild_boost_tier_2                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost_tier_3                        | guild boost tier 3                        | 11    | convert_guild_boost_tier_3                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | new_follower_channel                      | new follower channel                      | 12    | convert_new_follower_channel                      | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stream                                    | stream                                    | 13    | convert_stream                                    | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_disqualified                    | discovery disqualified                    | 14    | convert_discovery_disqualified                    | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_requalified                     | discovery requalified                     | 15    | convert_discovery_requalified                     | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_grace_period_initial_warning    | discovery grace period initial warning    | 16    | convert_discovery_grace_period_initial_warning    | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_grace_period_final_warning      | discovery grace period final warning      | 17    | convert_discovery_grace_period_final_warning      | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | thread_created                            | thread created                            | 18    | convert_thread_created                            | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | inline_reply                              | inline reply                              | 19    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | slash_command                             | slash command                             | 20    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | thread_started                            | thread started                            | 21    | MESSAGE_DEFAULT_CONVERTER                         | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | invite_reminder                           | invite reminder                           | 22    | convert_invite_reminder                           | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | context_command                           | context command                           | 23    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | auto_moderation_action                    | auto moderation action                    | 24    | convert_auto_moderation_action                    | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | role_subscription_purchase                | role subscription purchase                | 25    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | interaction_premium_upsell                | interaction premium upsell                | 26    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_start                               | stage start                               | 27    | convert_stage_start                               | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_end                                 | stage end                                 | 28    | convert_stage_end                                 | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_speaker                             | stage speaker                             | 29    | convert_stage_speaker                             | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_request_to_speak                    | stage request to speak                    | 30    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_topic_change                        | stage topic change                        | 31    | convert_stage_topic_change                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | application_subscription                  | application subscription                  | 32    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | private_channel_integration_add           | private channel integration add           | 33    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | private_channel_integration_remove        | private channel integration remove        | 34    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | premium_referral                          | premium referral                          | 35    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ('converter', 'deletable',)
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new message type with the given value.
        
        Parameters
        ----------
        value : `int`
            The message type's identifier value.
        
        Returns
        -------
        self : ``MessageType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.converter = MESSAGE_DEFAULT_CONVERTER
        self.deletable = True
        
        return self
    
    
    def __init__(self, value, name, converter, deletable):
        """
        Creates an ``MessageType`` and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the message type.
        name : `str`
            The default name of the message type.
        converter : `FunctionType`
            The converter function of the message type.
        deletable : `bool`
            Whether the message with this type can be deleted.
        """
        self.value = value
        self.name = name
        self.converter = converter
        self.deletable = deletable
        
        self.INSTANCES[value] = self
    
    
    def __repr__(self):
        """Returns the representation of the message type."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append(', value = ')
        repr_parts.append(repr(self.value))
        
        repr_parts.append(' converter = ')
        repr_parts.append(repr(self.converter))
        
        repr_parts.append(', deletable = ')
        repr_parts.append(repr(self.deletable))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    # predefined
    default = P(0, 'default', MESSAGE_DEFAULT_CONVERTER, True)
    user_add = P(1, 'add user', convert_user_add, False)
    user_remove = P(2, 'remove user', convert_user_remove, False)
    call = P(3, 'call', convert_call, False)
    channel_name_change = P(4, 'channel name change', convert_channel_name_change, False)
    channel_icon_change = P(5, 'channel icon change', convert_channel_icon_change, False)
    new_pin = P(6, 'new pin', convert_new_pin, True)
    welcome = P(7, 'welcome', convert_welcome, True)
    guild_boost = P(8, 'guild boost', convert_guild_boost, True)
    guild_boost_tier_1 = P(9, 'guild boost tier 1', convert_guild_boost_tier_1, True)
    guild_boost_tier_2 = P(10, 'guild boost tier 2', convert_guild_boost_tier_2, True)
    guild_boost_tier_3 = P(11, 'guild boost tier 3', convert_guild_boost_tier_3, True)
    new_follower_channel = P(12, 'new follower channel', convert_new_follower_channel, True)
    stream = P(13, 'stream', convert_stream, False)
    discovery_disqualified = P(14, 'discovery disqualified', convert_discovery_disqualified, False)
    discovery_requalified = P(15, 'discovery requalified', convert_discovery_requalified, False)
    discovery_grace_period_initial_warning = P(
        16, 'discovery grace period initial warning', convert_discovery_grace_period_initial_warning, False
    )
    discovery_grace_period_final_warning = P(
        17, 'discovery grace period final warning', convert_discovery_grace_period_final_warning, False
    )
    thread_created = P(18, 'thread created', convert_thread_created, True)
    inline_reply = P(19, 'inline reply', MESSAGE_DEFAULT_CONVERTER, True)
    slash_command = P(20, 'slash command', MESSAGE_DEFAULT_CONVERTER, True)
    thread_started = P(21, 'thread started', MESSAGE_DEFAULT_CONVERTER, False)
    invite_reminder = P(22, 'invite reminder', convert_invite_reminder, True)
    context_command = P(23, 'context command', MESSAGE_DEFAULT_CONVERTER, True)
    auto_moderation_action = P(24, 'auto moderation action', convert_auto_moderation_action, True)
    role_subscription_purchase = P (25, 'role subscription purchase', MESSAGE_DEFAULT_CONVERTER, True)
    interaction_premium_upsell = P(26, 'interaction premium upsell', MESSAGE_DEFAULT_CONVERTER, True)
    stage_start = P(27, 'stage start', convert_stage_start, True)
    stage_end = P(28, 'stage end', convert_stage_end, True)
    stage_speaker = P(29, 'stage speaker', convert_stage_speaker, True)
    stage_request_to_speak = P(30, 'stage request to speak', MESSAGE_DEFAULT_CONVERTER, True)
    stage_topic_change = P(31, 'stage topic change', convert_stage_topic_change, True)
    application_subscription = P(32, 'application subscription', MESSAGE_DEFAULT_CONVERTER, True)
    private_channel_integration_add = P(33, 'private channel integration add', MESSAGE_DEFAULT_CONVERTER, True)
    private_channel_integration_remove = P(34, 'private channel integration remove', MESSAGE_DEFAULT_CONVERTER, True)
    premium_referral = P(34, 'premium referral', MESSAGE_DEFAULT_CONVERTER, True)


    @class_property
    def add_user(cls):
        """
        Deprecated and will be removed in 2023 November. Please use `.user_add` instead.
        """
        warnings.warn(
            (
                f'`{cls.__name__}.add_user` is deprecated and will be removed in 2023 November. '
                f'Please use `.user_add` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return cls.user_add


    @class_property
    def remove_user(cls):
        """
        Deprecated and will be removed in 2023 November. Please use `.user_remove` instead.
        """
        warnings.warn(
            (
                f'`{cls.__name__}.remove_user` is deprecated and will be removed in 2023 November. '
                f'Please use `.user_remove` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return cls.user_remove


del convert_user_add
del convert_user_remove
del convert_call
del convert_channel_name_change
del convert_channel_icon_change
del convert_new_pin
del convert_welcome
del convert_guild_boost
del convert_guild_boost_tier_1
del convert_guild_boost_tier_2
del convert_guild_boost_tier_3
del convert_new_follower_channel
del convert_stream
del convert_discovery_disqualified
del convert_discovery_requalified
del convert_discovery_grace_period_initial_warning
del convert_discovery_grace_period_final_warning
del convert_thread_created
del convert_invite_reminder
del convert_stage_start
del convert_stage_end
del convert_stage_topic_change
del convert_stage_speaker
del convert_auto_moderation_action

GENERIC_MESSAGE_TYPES = frozenset((
    MessageType.default,
    MessageType.inline_reply,
    MessageType.slash_command,
    MessageType.thread_started,
    MessageType.context_command,
))



__all__ = ()

from ...channel.channel.fields import parse_id, put_id_into, validate_id
from ...channel.channel_metadata.fields import (
    parse_name, parse_position, put_name_into, put_position_into, validate_name, validate_position
)
